/*
	BorderControl plugin for Avisynth -- lets you manipulate
	the frame borders to diguise poor source material defects
	such as VCR head-switching noise

	Copyright (C) 2002-2003 Simon Walters, All Rights Reserved

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	The author can be contacted at:
	Simon Walters
	siwalters@hotmail.com

	V1.0 - 1st release.
	V1.1 - Add speed improvements
			Added a different mode for smear factor - just added for top
			and bottom at the moment.
	preV1.2
	V1.3 Works with Avisynth 2.5
	V1.4 Attempt to work in YV12
	V1.4a - Fix bug with black border being green - thanks to ivo
*/
#include "windows.h"
#include "avisynth.h"
//#include "internal.h"

#include "memory.h"



class BorderControl : public GenericVideoFilter {
	int YBotBorder, YBotFade, YBotSmear, YBotStretchFactor;
	int YTopBorder, YTopFade, YTopSmear, YTopStretchFactor;
	int XLeftBorder, XLeftFade, XLeftSmear, XLeftStretchFactor;
	int XRightBorder, XRightFade, XRightSmear, XRightStretchFactor;

public:
	BorderControl(PClip _child,
		int _YBotBorder, int _YBotFade, int _YBotSmear,int _YBotStretchFactor,
		int _YTopBorder, int _YTopFade, int _YTopSmear,int _YTopStretchFactor,
		int _XLeftBorder, int _XLeftFade, int _XLeftSmear,int _XLeftStretchFactor,
		int _XRightBorder, int _XRightFade, int _XRightSmear,int _XRightStretchFactor,
		IScriptEnvironment* env) :
	GenericVideoFilter(_child),
		YBotBorder(_YBotBorder), YBotFade(_YBotFade),
		 YBotSmear(_YBotSmear), YBotStretchFactor(_YBotStretchFactor),
		YTopBorder(_YTopBorder), YTopFade(_YTopFade),
		 YTopSmear(_YTopSmear), YTopStretchFactor(_YTopStretchFactor),
		XLeftBorder(_XLeftBorder), XLeftFade(_XLeftFade),
		 XLeftSmear(_XLeftSmear), XLeftStretchFactor(_XLeftStretchFactor),
		XRightBorder(_XRightBorder), XRightFade(_XRightFade),
		 XRightSmear(_XRightSmear), XRightStretchFactor(_XRightStretchFactor)

	{
	// parameter checking goes in here
//	if (vi.IsYV12())
//			env->ThrowError("BorderControl: input to filter must be in YUY2 or RGB");
	}

//    BorderControl(PClip _child) : GenericVideoFilter(_child) {}
    PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
};



PVideoFrame __stdcall BorderControl::GetFrame(int n, IScriptEnvironment* env) {
	
	if (!vi.IsYV12())
	{
		
		
		PVideoFrame src = child->GetFrame(n, env);
		PVideoFrame dst = env->NewVideoFrame(vi);
		
		const unsigned char* srcp = src->GetReadPtr();
		unsigned char* dstp = dst->GetWritePtr();
		
		
		const int dst_pitch = dst->GetPitch();
		const int dst_width = dst->GetRowSize();
		const int dst_height = dst->GetHeight();
		const int src_pitch = src->GetPitch();
		const int src_width = src->GetRowSize();
		const int src_height = src->GetHeight();
		
		int dst_modulo;
		dst_modulo = dst_pitch - dst_width;
		
		
		int w, h;
		
		//Copy src to dst
		srcp = src->GetReadPtr();
		dstp = dst->GetWritePtr();
		
		h = dst_height;
		do {
			memcpy(dstp,srcp,dst_width);
			srcp = srcp + src_pitch;
			dstp = dstp + dst_pitch;
		} while(--h);
		// end copy src to dst
		
		unsigned char oldY1,oldU,oldY2,oldV,newY1,newU,newY2,newV;
		
		//Top
		srcp = src->GetReadPtr();
		
		dstp = dst->GetWritePtr();
		
		
		int hMax=YTopBorder;
		if ((YTopSmear+YTopStretchFactor) > hMax) hMax = (YTopSmear+YTopStretchFactor);
		if (YTopFade > hMax) hMax = YTopFade;
		h = 1;
		if (hMax != 0) {
			do {
				w = 1;
				do {
					
					oldY1 = *dstp;
					oldU = *(dstp+1);
					oldY2 = *(dstp+2);
					oldV = *(dstp+3);
					newY1 = oldY1;
					newU = oldU;
					newY2 = oldY2;
					newV = oldV;
					
					
					if (h <= YTopBorder) {
						newY1 = 16;
						newY2 = 16;
						newU = 128;
						newV = 128;                                                                                                                                                                                                                                                                                      
					}
					
					if ((h <= (YTopSmear+YTopStretchFactor)) && (h > YTopBorder)) {
						float SFF = (float)(h - YTopBorder)/(float)(YTopSmear +YTopStretchFactor - YTopBorder);
						newY1 = (unsigned char)((*(srcp + src_pitch*(YTopSmear+YTopStretchFactor-h+1))
							- *(srcp + src_pitch*(YTopSmear-h+1))) * SFF)
							+ *(srcp + src_pitch*(YTopSmear-h+1));
						newY2 = (unsigned char)((*(2 + srcp + src_pitch*(YTopSmear+YTopStretchFactor-h+1))
							- *(2 + srcp + src_pitch*(YTopSmear-h+1))) * SFF)
							+ *(2 + srcp + src_pitch*(YTopSmear-h+1));
						
						newU = (unsigned char)((*(1 + srcp + src_pitch*(YTopSmear+YTopStretchFactor-h+1))
							- *(1 + srcp + src_pitch*(YTopSmear-h+1))) * SFF)
							+ *(1 + srcp + src_pitch*(YTopSmear-h+1));
						
						newV = (unsigned char)((*(3 + srcp + src_pitch*(YTopSmear+YTopStretchFactor-h+1))
							- *(3 + srcp + src_pitch*(YTopSmear-h+1))) * SFF)
							+ *(3 + srcp + src_pitch*(YTopSmear-h+1));
					}	
					
					if ((h > YTopBorder) && (h <= YTopFade)) {
						newY1 = ((newY1 - 16) * (h - YTopBorder)/(YTopFade - YTopBorder)) + 16;
						newY2 = ((newY2 - 16) * (h - YTopBorder)/(YTopFade - YTopBorder)) + 16;
					}
					*dstp = newY1;
					*(dstp+1) = newU;
					*(dstp+2) = newY2;
					*(dstp+3) = newV;
					srcp = srcp + 4;
					dstp = dstp + 4;
					w++;
				} while(w <= (dst_width / 4));
				srcp = srcp + dst_modulo;
				dstp = dstp + dst_modulo;
				h++;
			} while(h <= hMax); //end top
		}
		
		
		//Bottom
		srcp = src->GetReadPtr();
		srcp = srcp + (dst_pitch * dst_height) - dst_modulo;
		
		dstp = dst->GetWritePtr();
		dstp = dstp + (dst_pitch * dst_height) - dst_modulo;
		
		srcp = srcp - 4;
		dstp = dstp - 4;
		
		hMax=YBotBorder;
		if ((YBotSmear+YBotStretchFactor) > hMax) hMax = (YBotSmear+YBotStretchFactor);
		if (YBotFade > hMax) hMax = YBotFade;
		h = 1;
		if (hMax != 0) {
			do {
				w = 1;
				do {
					
					oldY1 = *dstp;
					oldU = *(dstp+1);
					oldY2 = *(dstp+2);
					oldV = *(dstp+3);
					newY1 = oldY1;
					newU = oldU;
					newY2 = oldY2;
					newV = oldV;
					
					
					if (h <= YBotBorder) {
						newY1 = 16;
						newY2 = 16;
						newU = 128;
						newV = 128;                                                                                                                                                                                                                                                                                      
					}
					
					if ((h <= (YBotSmear+YBotStretchFactor)) && (h > YBotBorder)) {
						float SFF = (float)(h - YBotBorder)/(float)(YBotSmear +YBotStretchFactor - YBotBorder);
						newY1 = (unsigned char)((*(srcp - src_pitch*(YBotSmear+YBotStretchFactor-h+1))
							- *(srcp - src_pitch*(YBotSmear-h+1))) * SFF)
							+ *(srcp - src_pitch*(YBotSmear-h+1));
						newY2 = (unsigned char)((*(2 + srcp - src_pitch*(YBotSmear+YBotStretchFactor-h+1))
							- *(2 + srcp - src_pitch*(YBotSmear-h+1))) * SFF)
							+ *(2 + srcp - src_pitch*(YBotSmear-h+1));
						
						newU = (unsigned char)((*(1 + srcp - src_pitch*(YBotSmear+YBotStretchFactor-h+1))
							- *(1 + srcp - src_pitch*(YBotSmear-h+1))) * SFF)
							+ *(1 + srcp - src_pitch*(YBotSmear-h+1));
						
						newV = (unsigned char)((*(3 + srcp - src_pitch*(YBotSmear+YBotStretchFactor-h+1))
							- *(3 + srcp - src_pitch*(YBotSmear-h+1))) * SFF)
							+ *(3 + srcp - src_pitch*(YBotSmear-h+1));
					}			
					
					if ((h > YBotBorder) && (h < YBotFade)) {
						newY1 = ((newY1 - 16) * (h - YBotBorder)/(YBotFade - YBotBorder)) + 16;
						newY2 = ((newY2 - 16) * (h - YBotBorder)/(YBotFade - YBotBorder)) + 16;
					}
					
					srcp = srcp - 4;
					
					*dstp = newY1;
					*(dstp+1) = newU;
					*(dstp+2) = newY2;
					*(dstp+3) = newV;
					
					dstp = dstp -4;
					w++;
				} while(w <= (dst_width / 4));
				srcp = srcp - dst_modulo;
				dstp = dstp - dst_modulo;
				h++;
			} while(h <= hMax);
		} //end bottom
		
		
		//Left
		srcp = src->GetReadPtr();
		dstp = dst->GetWritePtr();
		
		
		int wMax=XLeftBorder;
		if ((XLeftSmear+XLeftStretchFactor) > wMax) wMax = (XLeftSmear+XLeftStretchFactor);
		if (XLeftFade > hMax) wMax = XLeftFade;
		w = 1;
		if (wMax != 0) {
			
			do {
				h = 1;
				do {
					
					oldY1 = *dstp;
					oldU = *(dstp+1);
					oldY2 = *(dstp+2);
					oldV = *(dstp+3);
					newY1 = oldY1;
					newU = oldU;
					newY2 = oldY2;
					newV = oldV;
					
					
					if (w <= XLeftBorder) {
						newY1 = 16;
						newY2 = 16;
						newU = 128;
						newV = 128;                                                                                                                                                                                                                                                                                      
					}
					
					if ((w <= (XLeftSmear+XLeftStretchFactor)) && (w > XLeftBorder)) {
						int SFF = (XLeftStretchFactor + 1)*(w - XLeftBorder - 1)/
							(XLeftSmear + XLeftStretchFactor - XLeftBorder);
						newY1 = *(srcp + 4*(XLeftSmear - w + 1+ (int)(SFF)));
						newY2 = *(srcp + 4*(XLeftSmear - w + 1+ (int)(SFF)) + 2);
						newY2 = newY1;
						
						newU = *(srcp + 4*(XLeftSmear - w + 1+ (int)(SFF)) + 1);
						newV = *(srcp + 4*(XLeftSmear - w + 1+ (int)(SFF)) + 3);
					}			
					
					if ((w > XLeftBorder) && (w < XLeftFade)) {
						newY1 = ((newY1 - 16) * (w - XLeftBorder)/(XLeftFade - XLeftBorder)) + 16;
						newY2 = ((newY2 - 16) * (w - XLeftBorder)/(XLeftFade - XLeftBorder)) + 16;
					}
					
					
					*dstp = newY1;
					*(dstp+1) = newU;
					*(dstp+2) = newY2;
					*(dstp+3) = newV;
					
					srcp = srcp + src_pitch;
					dstp = dstp + src_pitch;
					h++;
				} while(h <= dst_height);
				srcp = src->GetReadPtr() + (w*4);
				dstp = dst->GetWritePtr() + (w*4);
				w++;
				
			} while(w <= (wMax));
		}

		// End Left
		
		//Right
		srcp = src->GetReadPtr();
		dstp = dst->GetWritePtr();
		
		srcp = srcp + src_width - 4;
		dstp = dstp + src_width - 4;
		
		wMax=XRightBorder;
		if ((XRightSmear+XRightStretchFactor) > wMax) wMax = (XRightSmear+XRightStretchFactor);
		if (XRightFade > hMax) wMax = XRightFade;
		w = 1;
		if (wMax != 0) {
			do {
				h = 1;
				do {
					oldY1 = *dstp;
					oldU = *(dstp+1);
					oldY2 = *(dstp+2);
					oldV = *(dstp+3);
					newY1 = oldY1;
					newU = oldU;
					newY2 = oldY2;
					newV = oldV;
					
					if (w <= XRightBorder) {
						newY1 = 16;
						newY2 = 16;
						newU = 128;
						newV = 128;                                                                                                                                                                                                                                                                                      
					}
					
					if ((w <= (XRightSmear+XRightStretchFactor)) && (w > XRightBorder)) {
						int SFF = (XRightStretchFactor + 1)*(w - XRightBorder - 1)/
							(XRightSmear + XRightStretchFactor - XRightBorder);
						newY1 = *(srcp - 4*(XRightSmear - w + 1+ (int)(SFF)));
						newY2 = *(srcp - 4*(XRightSmear - w + 1+ (int)(SFF)) + 2);
						newY2 = newY1;
						
						newU = *(srcp - 4*(XRightSmear - w + 1+ (int)(SFF)) + 1);
						newV = *(srcp - 4*(XRightSmear - w + 1+ (int)(SFF)) + 3);
					}			
					
					if ((w > XRightBorder) && (w < XRightFade)) {
						newY1 = ((newY1 - 16) * (w - XRightBorder)/(XRightFade - XRightBorder)) + 16;
						newY2 = ((newY2 - 16) * (w - XRightBorder)/(XRightFade - XRightBorder)) + 16;
					}
					
					
					*dstp = newY1;
					*(dstp+1) = newU;
					*(dstp+2) = newY2;
					*(dstp+3) = newV;
					
					srcp = srcp + src_pitch;
					dstp = dstp + src_pitch;
					h++;
				} while(h <= dst_height);
				srcp = src->GetReadPtr() + src_width - 4*(w + 1);
				dstp = dst->GetWritePtr() + src_width - 4*(w + 1);
				w++;
				
			} while(w <= (wMax));
		}
		// End Right
		
		
		return dst;
	}
	else
////////////////////////////////////////////////////////////////////////////
	// Code to handle YV12

	{
		PVideoFrame src = child->GetFrame(n, env);
		PVideoFrame dst = env->NewVideoFrame(vi);
		// Y Section		
		const unsigned char* srcp = src->GetReadPtr(PLANAR_Y);
		unsigned char* dstp = dst->GetWritePtr(PLANAR_Y);
		
		
		int dst_pitch = dst->GetPitch(PLANAR_Y);
		int dst_width = dst->GetRowSize(PLANAR_Y);
		int dst_height = dst->GetHeight(PLANAR_Y);
		int src_pitch = src->GetPitch(PLANAR_Y);
		int src_width = src->GetRowSize(PLANAR_Y);
		int src_height = src->GetHeight(PLANAR_Y);
		
		int dst_modulo;
		dst_modulo = dst_pitch - dst_width;
		int wMax;
		
		
		int w, h;
		
		//Copy Y src to dst
		int PLANE=PLANAR_Y;
		srcp = src->GetReadPtr(PLANE);
		dstp = dst->GetWritePtr(PLANE);
		
		h = dst_height;
		do {
			memcpy(dstp,srcp,dst_width);
			srcp = srcp + src_pitch;
			dstp = dstp + dst_pitch;
		} while(--h);
		// end copy src to dst


		
		//Top
		srcp = src->GetReadPtr(PLANE);
		dstp = dst->GetWritePtr(PLANE);
		
		
		int hMax=YTopBorder;
		if ((YTopSmear+YTopStretchFactor) > hMax) hMax = (YTopSmear+YTopStretchFactor);
		if (YTopFade > hMax) hMax = YTopFade;
		h = 1;
		if (hMax != 0) {
			do {
				w = 1;
				do {
					
					if (h <= YTopBorder) {
						*dstp = 16;
					}
					
					if ((h <= (YTopSmear+YTopStretchFactor)) && (h > YTopBorder)) {
						float SFF = (float)(h - YTopBorder)/(float)(YTopSmear +YTopStretchFactor - YTopBorder);
						*dstp = (unsigned char)((*(srcp + src_pitch*(YTopSmear+YTopStretchFactor-h+1))
							- *(srcp + src_pitch*(YTopSmear-h+1))) * SFF)
							+ *(srcp + src_pitch*(YTopSmear-h+1));
					}	
					
					if ((h > YTopBorder) && (h <= YTopFade)) {
						*dstp = ((*dstp - 16) * (h - YTopBorder)/(YTopFade - YTopBorder)) + 16;
					}

					srcp = srcp + 1;
					dstp = dstp + 1;
					w++;
				} while(w <= (dst_width));
				srcp = srcp + dst_modulo;
				dstp = dstp + dst_modulo;
				h++;
			} while(h <= hMax); //end top
		}
		
		
		//Bottom
		srcp = src->GetReadPtr(PLANE);
		dstp = dst->GetWritePtr(PLANE);

		srcp = srcp + (dst_pitch * dst_height) - dst_modulo;
		dstp = dstp + (dst_pitch * dst_height) - dst_modulo;
		
		srcp = srcp - 1;
		dstp = dstp - 1;
		
		hMax=YBotBorder;
		if ((YBotSmear+YBotStretchFactor) > hMax) hMax = (YBotSmear+YBotStretchFactor);
		if (YBotFade > hMax) hMax = YBotFade;
		h = 1;
		if (hMax != 0) {
			do {
				w = 1;
				do {
					
					if (h <= YBotBorder) {
						*dstp = 16;
					}
					
					if ((h <= (YBotSmear+YBotStretchFactor)) && (h > YBotBorder)) {
						float SFF = (float)(h - YBotBorder)/(float)(YBotSmear +YBotStretchFactor - YBotBorder);
						*dstp = (unsigned char)((*(srcp - src_pitch*(YBotSmear+YBotStretchFactor-h+1))
							- *(srcp - src_pitch*(YBotSmear-h+1))) * SFF)
							+ *(srcp - src_pitch*(YBotSmear-h+1));
					}			
					
					if ((h > YBotBorder) && (h < YBotFade)) {
						*dstp = ((*dstp - 16) * (h - YBotBorder)/(YBotFade - YBotBorder)) + 16;
					}
					
					srcp = srcp - 1;
					dstp = dstp - 1;
					w++;
				} while(w <= (dst_width));
				srcp = srcp - dst_modulo;
				dstp = dstp - dst_modulo;
				h++;
			} while(h <= hMax);
		} //end bottom
		
		
		//Left
		srcp = src->GetReadPtr(PLANE);
		dstp = dst->GetWritePtr(PLANE);
		
		
		wMax=XLeftBorder;
		if ((XLeftSmear+XLeftStretchFactor) > wMax) wMax = (XLeftSmear+XLeftStretchFactor);
		if (XLeftFade > hMax) wMax = XLeftFade;
		w = 1;
		if (wMax != 0) {
			
			do {
				h = 1;
				do {
					
					if (w <= XLeftBorder) {
						*dstp = 16;
					}
					
					if ((w <= (XLeftSmear+XLeftStretchFactor)) && (w > XLeftBorder)) {
						int SFF = (XLeftStretchFactor + 1)*(w - XLeftBorder - 1)/
							(XLeftSmear + XLeftStretchFactor - XLeftBorder);
						*dstp = *(srcp + (XLeftSmear - w + 1+ (int)(SFF)));
					}			
					
					if ((w > XLeftBorder) && (w < XLeftFade)) {
						*dstp = ((*dstp - 16) * (w - XLeftBorder)/(XLeftFade - XLeftBorder)) + 16;
					}
					
					
					srcp = srcp + src_pitch;
					dstp = dstp + src_pitch;
					h++;
				} while(h <= dst_height);
				srcp = src->GetReadPtr(PLANE) + (w);
				dstp = dst->GetWritePtr(PLANE) + (w);
				w++;
				
			} while(w <= (wMax));
		}

		// End Left
		
		//Right
		srcp = src->GetReadPtr(PLANE);
		dstp = dst->GetWritePtr(PLANE);
		
		srcp = srcp + src_width - 1;
		dstp = dstp + src_width - 1;
		
		wMax=XRightBorder;
		if ((XRightSmear+XRightStretchFactor) > wMax) wMax = (XRightSmear+XRightStretchFactor);
		if (XRightFade > hMax) wMax = XRightFade;
		w = 1;
		if (wMax != 0) {
			do {
				h = 1;
				do {
			
					if (w <= XRightBorder) {
						*dstp = 16;
                                                                                                                                                                                                                                                                                   
					}
					
					if ((w <= (XRightSmear+XRightStretchFactor)) && (w > XRightBorder)) {
						int SFF = (XRightStretchFactor + 1)*(w - XRightBorder - 1)/
							(XRightSmear + XRightStretchFactor - XRightBorder);
						*dstp = *(srcp - (XRightSmear - w + 1+ (int)(SFF)));
					}			
					
					if ((w > XRightBorder) && (w < XRightFade)) {
						*dstp = ((*dstp - 16) * (w - XRightBorder)/(XRightFade - XRightBorder)) + 16;
					}
					
					srcp = srcp + src_pitch;
					dstp = dstp + src_pitch;
					h++;
				} while(h <= dst_height);
				srcp = src->GetReadPtr(PLANE) + src_width - (w + 1);
				dstp = dst->GetWritePtr(PLANE) + src_width - (w + 1);
				w++;
				
			} while(w <= (wMax));
		}
		// End Right
		

		int YTBold= YTopBorder;
		int YTSold= YTopSmear;
		int YTFold= YTopFade;
		int YTSFold= YTopStretchFactor;

		int YBBold= YBotBorder;
		int YBSold= YBotSmear;
		int YBFold= YBotFade;
		int YBSFold= YBotStretchFactor;

		int XLBold= XLeftBorder;
		int XLSold= XLeftSmear;
		int XLFold= XLeftFade;
		int XLSFold= XLeftStretchFactor;


		int XRBold= XRightBorder;
		int XRSold= XRightSmear;
		int XRFold= XRightFade;
		int XRSFold= XRightStretchFactor;

		YTopBorder >>= 1;
		YTopSmear >>= 1;
		YTopFade >>= 1;
		YTopStretchFactor >>= 1;

		YBotBorder >>= 1;
		YBotSmear >>= 1;
		YBotFade >>= 1;
		YBotStretchFactor >>= 1;

		XLeftBorder >>= 1;
		XLeftSmear >>= 1;
		XLeftFade >>= 1;
		XLeftStretchFactor >>= 1;

		XRightBorder >>= 1;
		XRightSmear >>= 1;
		XRightFade >>= 1;
		XRightStretchFactor >>= 1;


////// start UV Code section		
		dst_pitch = dst->GetPitch(PLANAR_U);
		dst_width = dst->GetRowSize(PLANAR_U);
		dst_height = dst->GetHeight(PLANAR_U);
		src_pitch = src->GetPitch(PLANAR_U);
		src_width = src->GetRowSize(PLANAR_U);
		src_height = src->GetHeight(PLANAR_U);
		
		dst_modulo;
		dst_modulo = dst_pitch - dst_width;

////// Start U section		
		PLANE=PLANAR_U;
		//Copy src to dst
		srcp = src->GetReadPtr(PLANE);
		dstp = dst->GetWritePtr(PLANE);
		
		h = dst_height;
		do {
			memcpy(dstp,srcp,dst_width);
			srcp = srcp + src_pitch;
			dstp = dstp + dst_pitch;
		} while(--h);
		// end copy src to dst

		
		//Top
		srcp = src->GetReadPtr(PLANE);
		dstp = dst->GetWritePtr(PLANE);
		
		
		hMax=YTopBorder;
		if ((YTopSmear+YTopStretchFactor) > hMax) hMax = (YTopSmear+YTopStretchFactor);
		if (YTopFade > hMax) hMax = YTopFade;
		h = 1;
		if (hMax != 0) {
			do {
				w = 1;
				do {
					
					if (h <= YTopBorder) {
						*dstp = 128;
					}
					
					if ((h <= (YTopSmear+YTopStretchFactor)) && (h > YTopBorder)) {
						float SFF = (float)(h - YTopBorder)/(float)(YTopSmear +YTopStretchFactor - YTopBorder);
						*dstp = (unsigned char)((*(srcp + src_pitch*(YTopSmear+YTopStretchFactor-h+1))
							- *(srcp + src_pitch*(YTopSmear-h+1))) * SFF)
							+ *(srcp + src_pitch*(YTopSmear-h+1));
					}	
					
					if ((h > YTopBorder) && (h <= YTopFade)) {
						*dstp = ((*dstp - 16) * (h - YTopBorder)/(YTopFade - YTopBorder)) + 16;
					}

					srcp = srcp + 1;
					dstp = dstp + 1;
					w++;
				} while(w <= (dst_width));
				srcp = srcp + dst_modulo;
				dstp = dstp + dst_modulo;
				h++;
			} while(h <= hMax); //end top
		}

		//Bottom
		srcp = src->GetReadPtr(PLANE);
		dstp = dst->GetWritePtr(PLANE);

		srcp = srcp + (dst_pitch * dst_height) - dst_modulo;
		dstp = dstp + (dst_pitch * dst_height) - dst_modulo;
		
		srcp = srcp - 1;
		dstp = dstp - 1;
		
		hMax=YBotBorder;
		if ((YBotSmear+YBotStretchFactor) > hMax) hMax = (YBotSmear+YBotStretchFactor);
		if (YBotFade > hMax) hMax = YBotFade;
		h = 1;
		if (hMax != 0) {
			do {
				w = 1;
				do {
					
					if (h <= YBotBorder) {
						*dstp = 128;
					}
					
					if ((h <= (YBotSmear+YBotStretchFactor)) && (h > YBotBorder)) {
						float SFF = (float)(h - YBotBorder)/(float)(YBotSmear +YBotStretchFactor - YBotBorder);
						*dstp = (unsigned char)((*(srcp - src_pitch*(YBotSmear+YBotStretchFactor-h+1))
							- *(srcp - src_pitch*(YBotSmear-h+1))) * SFF)
							+ *(srcp - src_pitch*(YBotSmear-h+1));
					}			
					
					if ((h > YBotBorder) && (h < YBotFade)) {
						*dstp = ((*dstp - 16) * (h - YBotBorder)/(YBotFade - YBotBorder)) + 16;
					}
					
					srcp = srcp - 1;
					dstp = dstp - 1;
					w++;
				} while(w <= (dst_width));
				srcp = srcp - dst_modulo;
				dstp = dstp - dst_modulo;
				h++;
			} while(h <= hMax);
		} //end bottom
		//Left
		srcp = src->GetReadPtr(PLANE);
		dstp = dst->GetWritePtr(PLANE);
		
		
		wMax=XLeftBorder;
		if ((XLeftSmear+XLeftStretchFactor) > wMax) wMax = (XLeftSmear+XLeftStretchFactor);
		if (XLeftFade > hMax) wMax = XLeftFade;
		w = 1;
		if (wMax != 0) {
			
			do {
				h = 1;
				do {
					
					if (w <= XLeftBorder) {
						*dstp = 128;
					}
					
					if ((w <= (XLeftSmear+XLeftStretchFactor)) && (w > XLeftBorder)) {
						int SFF = (XLeftStretchFactor + 1)*(w - XLeftBorder - 1)/
							(XLeftSmear + XLeftStretchFactor - XLeftBorder);
						*dstp = *(srcp + (XLeftSmear - w + 1+ (int)(SFF)));
					}			
					
					if ((w > XLeftBorder) && (w < XLeftFade)) {
						*dstp = ((*dstp - 16) * (w - XLeftBorder)/(XLeftFade - XLeftBorder)) + 16;
					}
					
					
					srcp = srcp + src_pitch;
					dstp = dstp + src_pitch;
					h++;
				} while(h <= dst_height);
				srcp = src->GetReadPtr(PLANE) + (w);
				dstp = dst->GetWritePtr(PLANE) + (w);
				w++;
				
			} while(w <= (wMax));
		}

		// End Left
		//Right
		srcp = src->GetReadPtr(PLANE);
		dstp = dst->GetWritePtr(PLANE);
		
		srcp = srcp + src_width - 1;
		dstp = dstp + src_width - 1;
		
		wMax=XRightBorder;
		if ((XRightSmear+XRightStretchFactor) > wMax) wMax = (XRightSmear+XRightStretchFactor);
		if (XRightFade > hMax) wMax = XRightFade;
		w = 1;
		if (wMax != 0) {
			do {
				h = 1;
				do {
			
					if (w <= XRightBorder) {
						*dstp = 128;
                                                                                                                                                                                                                                                                                   
					}
					
					if ((w <= (XRightSmear+XRightStretchFactor)) && (w > XRightBorder)) {
						int SFF = (XRightStretchFactor + 1)*(w - XRightBorder - 1)/
							(XRightSmear + XRightStretchFactor - XRightBorder);
						*dstp = *(srcp - (XRightSmear - w + 1+ (int)(SFF)));
					}			
					
					if ((w > XRightBorder) && (w < XRightFade)) {
						*dstp = ((*dstp - 16) * (w - XRightBorder)/(XRightFade - XRightBorder)) + 16;
					}
					
					srcp = srcp + src_pitch;
					dstp = dstp + src_pitch;
					h++;
				} while(h <= dst_height);
				srcp = src->GetReadPtr(PLANE) + src_width - (w + 1);
				dstp = dst->GetWritePtr(PLANE) + src_width - (w + 1);
				w++;
				
			} while(w <= (wMax));
		}
		// End Right

		
///end U section

////// Start of V section
		PLANE=PLANAR_V;
		//Copy src to dst
		srcp = src->GetReadPtr(PLANE);
		dstp = dst->GetWritePtr(PLANE);
		
		h = dst_height;
		do {
			memcpy(dstp,srcp,dst_width);
			srcp = srcp + src_pitch;
			dstp = dstp + dst_pitch;
		} while(--h);
		// end copy src to dst

		
		//Top
		srcp = src->GetReadPtr(PLANE);
		dstp = dst->GetWritePtr(PLANE);
		
		
		hMax=YTopBorder;
		if ((YTopSmear+YTopStretchFactor) > hMax) hMax = (YTopSmear+YTopStretchFactor);
		if (YTopFade > hMax) hMax = YTopFade;
		h = 1;
		if (hMax != 0) {
			do {
				w = 1;
				do {
					
					if (h <= YTopBorder) {
						*dstp = 128;
					}
					
					if ((h <= (YTopSmear+YTopStretchFactor)) && (h > YTopBorder)) {
						float SFF = (float)(h - YTopBorder)/(float)(YTopSmear +YTopStretchFactor - YTopBorder);
						*dstp = (unsigned char)((*(srcp + src_pitch*(YTopSmear+YTopStretchFactor-h+1))
							- *(srcp + src_pitch*(YTopSmear-h+1))) * SFF)
							+ *(srcp + src_pitch*(YTopSmear-h+1));
					}	
					
					if ((h > YTopBorder) && (h <= YTopFade)) {
						*dstp = ((*dstp - 16) * (h - YTopBorder)/(YTopFade - YTopBorder)) + 16;
					}

					srcp = srcp + 1;
					dstp = dstp + 1;
					w++;
				} while(w <= (dst_width));
				srcp = srcp + dst_modulo;
				dstp = dstp + dst_modulo;
				h++;
			} while(h <= hMax); //end top
		}

		//Bottom
		srcp = src->GetReadPtr(PLANE);
		dstp = dst->GetWritePtr(PLANE);

		srcp = srcp + (dst_pitch * dst_height) - dst_modulo;
		dstp = dstp + (dst_pitch * dst_height) - dst_modulo;
		
		srcp = srcp - 1;
		dstp = dstp - 1;
		
		hMax=YBotBorder;
		if ((YBotSmear+YBotStretchFactor) > hMax) hMax = (YBotSmear+YBotStretchFactor);
		if (YBotFade > hMax) hMax = YBotFade;
		h = 1;
		if (hMax != 0) {
			do {
				w = 1;
				do {
					
					if (h <= YBotBorder) {
						*dstp = 128;
					}
					
					if ((h <= (YBotSmear+YBotStretchFactor)) && (h > YBotBorder)) {
						float SFF = (float)(h - YBotBorder)/(float)(YBotSmear +YBotStretchFactor - YBotBorder);
						*dstp = (unsigned char)((*(srcp - src_pitch*(YBotSmear+YBotStretchFactor-h+1))
							- *(srcp - src_pitch*(YBotSmear-h+1))) * SFF)
							+ *(srcp - src_pitch*(YBotSmear-h+1));
					}			
					
					if ((h > YBotBorder) && (h < YBotFade)) {
						*dstp = ((*dstp - 16) * (h - YBotBorder)/(YBotFade - YBotBorder)) + 16;
					}
					
					srcp = srcp - 1;
					dstp = dstp - 1;
					w++;
				} while(w <= (dst_width));
				srcp = srcp - dst_modulo;
				dstp = dstp - dst_modulo;
				h++;
			} while(h <= hMax);
		} //end bottom

		//Left
		srcp = src->GetReadPtr(PLANE);
		dstp = dst->GetWritePtr(PLANE);
		
		
		wMax=XLeftBorder;
		if ((XLeftSmear+XLeftStretchFactor) > wMax) wMax = (XLeftSmear+XLeftStretchFactor);
		if (XLeftFade > hMax) wMax = XLeftFade;
		w = 1;
		if (wMax != 0) {
			
			do {
				h = 1;
				do {
					
					if (w <= XLeftBorder) {
						*dstp = 128;
					}
					
					if ((w <= (XLeftSmear+XLeftStretchFactor)) && (w > XLeftBorder)) {
						int SFF = (XLeftStretchFactor + 1)*(w - XLeftBorder - 1)/
							(XLeftSmear + XLeftStretchFactor - XLeftBorder);
						*dstp = *(srcp + (XLeftSmear - w + 1+ (int)(SFF)));
					}			
					
					if ((w > XLeftBorder) && (w < XLeftFade)) {
						*dstp = ((*dstp - 16) * (w - XLeftBorder)/(XLeftFade - XLeftBorder)) + 16;
					}
					
					
					srcp = srcp + src_pitch;
					dstp = dstp + src_pitch;
					h++;
				} while(h <= dst_height);
				srcp = src->GetReadPtr(PLANE) + (w);
				dstp = dst->GetWritePtr(PLANE) + (w);
				w++;
				
			} while(w <= (wMax));
		}

		// End Left
		//Right
		srcp = src->GetReadPtr(PLANE);
		dstp = dst->GetWritePtr(PLANE);
		
		srcp = srcp + src_width - 1;
		dstp = dstp + src_width - 1;
		
		wMax=XRightBorder;
		if ((XRightSmear+XRightStretchFactor) > wMax) wMax = (XRightSmear+XRightStretchFactor);
		if (XRightFade > hMax) wMax = XRightFade;
		w = 1;
		if (wMax != 0) {
			do {
				h = 1;
				do {
			
					if (w <= XRightBorder) {
						*dstp = 128;
                                                                                                                                                                                                                                                                                   
					}
					
					if ((w <= (XRightSmear+XRightStretchFactor)) && (w > XRightBorder)) {
						int SFF = (XRightStretchFactor + 1)*(w - XRightBorder - 1)/
							(XRightSmear + XRightStretchFactor - XRightBorder);
						*dstp = *(srcp - (XRightSmear - w + 1+ (int)(SFF)));
					}			
					
					if ((w > XRightBorder) && (w < XRightFade)) {
						*dstp = ((*dstp - 16) * (w - XRightBorder)/(XRightFade - XRightBorder)) + 16;
					}
					
					srcp = srcp + src_pitch;
					dstp = dstp + src_pitch;
					h++;
				} while(h <= dst_height);
				srcp = src->GetReadPtr(PLANE) + src_width - (w + 1);
				dstp = dst->GetWritePtr(PLANE) + src_width - (w + 1);
				w++;
				
			} while(w <= (wMax));
		}
		// End Right
		
///end V section
/*
		//Clear U dst
		dstp = dst->GetWritePtr(PLANAR_U);
		int max = max(max(max(YTBold,YTFold),YTSold),YTSFold) >> 1;
		
		h = dst_height;
		do {
//			memcpy(dstp,srcp,dst_width);
			w=dst_width;
			do {
				*dstp = 128;
				dstp++;
			} while (--w);
			dstp = dstp + dst_modulo;
		} while(--h);
		// end clear U Dst
		dstp = dst->GetWritePtr(PLANAR_V);
//		int max = YTBold || YTFold || YTSold || YTSFold;
		
		h = dst_height;
		do {
//			memcpy(dstp,srcp,dst_width);
			w=dst_width;
			do {
				*dstp = 128;
				dstp++;
			} while (--w);
			dstp = dstp + dst_modulo;
		} while(--h);		// end clear V Dst

		//Copy U src to dst
		srcp = src->GetReadPtr(PLANAR_U);
		dstp = dst->GetWritePtr(PLANAR_U);
		
		h = dst_height;
		do {
			memcpy(dstp,srcp,dst_width);
//			w=width;
//			do {
//				*dstp = 128;
//				dstp++;
//			} while (--w);
			srcp = srcp + src_pitch;
			dstp = dstp + dst_pitch;
		} while(--h);
		// end copy Usrc to dst

				//Copy V src to dst
		srcp = src->GetReadPtr(PLANAR_V);
		dstp = dst->GetWritePtr(PLANAR_V);
		
		h = dst_height;
		do {
			memcpy(dstp,srcp,dst_width);
//			w=width;
//			do {
//				*dstp = 128;
//				dstp++;
//			} while (--w);
			srcp = srcp + src_pitch;
			dstp = dstp + dst_pitch;
		} while(--h);
		// end copy V src to dst

*/


		YTopBorder	= YTBold;
		YTopSmear	= YTSold;
		YTopFade		= YTFold;
		YTopStretchFactor	= YTSFold;

		YBotBorder= YBBold;
		YBotSmear= YBSold;
		YBotFade= YBFold;
		YBotStretchFactor= YBSFold;

		XLeftBorder= XLBold;
		XLeftSmear= XLSold;
		XLeftFade= XLFold;
		XLeftStretchFactor= XLSFold;


		XRightBorder= XRBold;
		XRightSmear= XRSold;
		XRightFade= XRFold;
		XRightStretchFactor= XRSFold;





		
		return dst;
	}

	
	
}


AVSValue __cdecl Create_BorderControl(AVSValue args, void* user_data, IScriptEnvironment* env) {
    return new BorderControl(args[0].AsClip(),
		 args[1].AsInt(0),	// YBotBorder
		 args[2].AsInt(0),	// YBotFade
		 args[3].AsInt(0),	// YBotSmear
		 args[4].AsInt(0),	// YBotStretchFactor
		 args[5].AsInt(0),	// YTopBorder
		 args[6].AsInt(0),	// YTopFade
		 args[7].AsInt(0),	// YTopSmear
		 args[8].AsInt(0),	// YTopStretchFactor
		 args[9].AsInt(0),	// XLeftBorder
		 args[10].AsInt(0),	// XLeftFade
		 args[11].AsInt(0),	// XLeftSmear
		 args[12].AsInt(0),	// XLeftStretchFactor
		 args[13].AsInt(0),	// XRightBorder
		 args[14].AsInt(0),	// XRightFade
		 args[15].AsInt(0),	// XRightSmear
		 args[16].AsInt(0),	// XRightStretchFactor

		 env);
}


extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env) {
    env->AddFunction("BorderControl", "c[YBB]i[YBF]i[YBS]i[YBSF]i[YTB]i[YTF]i[YTS]i[YTSF]i[XLB]i[XLF]i[XLS]i[XLSF]i[XRB]i[XRF]i[XRS]i[XRSF]i",
		 Create_BorderControl, 0);
    return "`BorderControl' BorderControl plugin";
}

